# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


import sys
from datetime import timedelta
from unittest.mock import MagicMock, Mock, call

import pytest
from craft_providers import Base, ProviderError, bases, lxd
from craft_providers.lxd import LXDError
from freezegun import freeze_time
from logassert import Exact  # type: ignore


@pytest.fixture()
def mock_base_configuration():
    mock_base = Mock(spec=Base)
    mock_base.compatibility_tag = "mock-compat-tag-v200"
    mock_base.get_command_environment.return_value = {"foo": "bar"}
    return mock_base


@pytest.fixture()
def mock_lxc(mocker):
    _mock_lxc = mocker.patch("craft_providers.lxd.launcher.LXC", spec=lxd.LXC)
    _mock_lxc.return_value.project_list.return_value = ["default", "test-project"]
    return _mock_lxc.return_value


@pytest.fixture()
def mock_platform(mocker):
    mocker.patch("sys.platform", "linux")


@pytest.fixture()
def mock_timezone(fake_process):
    fake_process.register_subprocess(
        ["timedatectl", "show", "-p", "Timezone", "--value"],
        stdout="fake/timezone",
        occurrences=10,
    )


@pytest.fixture()
def fake_instance():
    """Returns a fake LXD Instance"""
    instance = MagicMock()
    # the name has an invalid character to ensure the instance_name will be different
    # so the two are not conflated in the unit tests
    instance.name = "test-instance-$"
    instance.instance_name = "test-instance-fa2d407652a1c51f6019"
    instance.project = "test-project"
    instance.remote = "test-remote"
    instance.exists.return_value = False
    instance.is_running.return_value = False
    return instance


@pytest.fixture()
def fake_base_instance(fake_process):
    """Returns a fake base LXD Instance"""
    base_instance = MagicMock()
    # the name has an invalid character to ensure the instance_name will be different,
    # so the two are not conflated in the unit tests
    base_instance.name = "test-base-instance-$"
    base_instance.instance_name = "test-base-instance-e14661a426076717fa04"
    base_instance.project = "test-project"
    base_instance.remote = "test-remote"
    base_instance.exists.return_value = False
    base_instance.is_running.return_value = False
    fake_process.register_subprocess(
        [
            "lxc",
            "--project",
            "test-project",
            "config",
            "get",
            "user.craft_providers.status",
            "test-remote:test-base-instance-e14661a426076717fa04",
        ],
        stdout="fake/timezone",
        occurrences=10,
    )
    return base_instance


@pytest.fixture()
def mock_lxd_instance(fake_instance, fake_base_instance, mocker):
    """Mock LXD instance to return fake_instance then fake_base_instance."""
    return mocker.patch(
        "craft_providers.lxd.launcher.LXDInstance",
        spec=lxd.LXDInstance,
        side_effect=[fake_instance, fake_base_instance],
    )


@pytest.fixture()
def mock_is_valid(mocker):
    return mocker.patch("craft_providers.lxd.launcher._is_valid", return_value=True)


@pytest.fixture()
def mock_check_id_map(mocker):
    return mocker.patch("craft_providers.lxd.launcher._check_id_map", return_value=True)


@pytest.fixture()
def mock_disable_timer_update_thread(mocker):
    return mocker.patch("craft_providers.lxd.launcher.InstanceTimer.run")


def test_launch_no_base_instance(
    fake_instance,
    mock_base_configuration,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
):
    """Create an instance from an image and do not save a copy as the base instance."""
    fake_instance.config_get.return_value = "STARTING"
    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            use_base_instance=False,
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("local"),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="default",
            remote="local",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=False,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.config_set("user.craft_providers.timer", "2023-01-01T00:00:00+00:00"),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="default",
            remote="local",
        ),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.stop(),
        call.start(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.setup(executor=fake_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_use_base_instance(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_is_valid,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
):
    """Launch a base instance from an image and copy to the new instance."""
    lxd.launch(
        name=fake_instance.name,
        base_configuration=mock_base_configuration,
        image_name="image-name",
        image_remote="image-remote",
        use_base_instance=True,
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
    )

    assert mock_lxc.mock_calls == [
        call.project_list("test-remote"),
        call.copy(
            source_remote="test-remote",
            source_instance_name=fake_base_instance.instance_name,
            destination_remote="test-remote",
            destination_instance_name=fake_instance.instance_name,
            project="test-project",
        ),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
        call(
            name="base-instance-mock-compat-tag-v200-image-remote-image-name",
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.start(),
    ]
    fake_base_instance.exists.assert_called_once()
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.get_command_environment(),
        call.setup(executor=fake_base_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_use_base_instance_failed_lxc(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_is_valid,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
):
    """Launch a base instance from an image, but lxc commands fail."""
    fake_base_instance.lxc.config_set.side_effect = [
        LXDError("test1"),
    ]
    with pytest.raises(LXDError):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            use_base_instance=True,
            project="test-project",
            remote="test-remote",
            lxc=mock_lxc,
        )


@pytest.mark.parametrize(("map_user_uid", "uid"), [(True, 1234), (False, None)])
def test_launch_use_existing_base_instance(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_is_valid,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
    map_user_uid,
    uid,
):
    """Create instance from an existing base instance."""
    fake_base_instance.exists.return_value = True
    mock_lxc.config_get.return_value = "both 1234 0"

    lxd.launch(
        name=fake_instance.name,
        base_configuration=mock_base_configuration,
        image_name="image-name",
        image_remote="image-remote",
        map_user_uid=map_user_uid,
        uid=uid,
        use_base_instance=True,
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
    )

    expected_mock_lxc_calls = [
        call.project_list("test-remote"),
        call.copy(
            source_remote="test-remote",
            source_instance_name=fake_base_instance.instance_name,
            destination_remote="test-remote",
            destination_instance_name=fake_instance.instance_name,
            project="test-project",
        ),
    ]
    if map_user_uid:
        expected_mock_lxc_calls.append(
            call.config_set(
                instance_name=fake_instance.instance_name,
                key="raw.idmap",
                value="both 1234 0",
                project="test-project",
                remote="test-remote",
            ),
        )
        expected_mock_lxc_calls.append(
            call.config_get(
                instance_name=fake_instance.instance_name,
                key="raw.idmap",
                project="test-project",
                remote="test-remote",
            ),
        )
    expected_mock_lxc_calls.append(
        call.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        )
    )
    assert mock_lxc.mock_calls == expected_mock_lxc_calls
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
        call(
            name="base-instance-mock-compat-tag-v200-image-remote-image-name",
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [call.exists(), call.is_running(), call.start()]
    assert fake_base_instance.mock_calls == [
        call.exists(),
        call.lxc.check_instance_status(
            instance_name=fake_base_instance.instance_name,
            project=fake_base_instance.project,
            remote=fake_base_instance.remote,
        ),
        call.is_running(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.get_command_environment(),
        call._setup_hostname(executor=fake_instance),
        call.warmup(executor=fake_instance),
    ]


def test_launch_use_existing_base_instance_already_running(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_is_valid,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
):
    """Launch an existing instance which is already running."""
    fake_base_instance.exists.return_value = True
    fake_base_instance.is_running.return_value = True

    fake_instance.is_running.return_value = True
    mock_lxc.config_get.return_value = "both 1234 0"

    lxd.launch(
        name=fake_instance.name,
        base_configuration=mock_base_configuration,
        image_name="image-name",
        image_remote="image-remote",
        map_user_uid=True,
        uid=1234,
        use_base_instance=True,
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
    )

    assert fake_instance.mock_calls == [
        call.exists(),
        call.is_running(),
        call.stop(),
        call.start(),
    ]
    assert fake_base_instance.mock_calls == [
        call.exists(),
        call.lxc.check_instance_status(
            instance_name=fake_base_instance.instance_name,
            project=fake_base_instance.project,
            remote=fake_base_instance.remote,
        ),
        call.is_running(),
    ]


def test_launch_existing_base_instance_invalid(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_is_valid,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
):
    """If the existing base instance is invalid, delete it and create a new instance."""
    fake_base_instance.exists.return_value = True
    mock_is_valid.return_value = False
    fake_base_instance.config_get.return_value = "STARTING"

    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            use_base_instance=True,
            project="test-project",
            remote="test-remote",
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("test-remote"),
        call.copy(
            source_remote="test-remote",
            source_instance_name=fake_base_instance.instance_name,
            destination_remote="test-remote",
            destination_instance_name=fake_instance.instance_name,
            project="test-project",
        ),
    ]

    assert fake_base_instance.lxc.mock_calls == [
        call.config_set(
            instance_name="test-base-instance-e14661a426076717fa04",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
    ]

    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
        call(
            name="base-instance-mock-compat-tag-v200-image-remote-image-name",
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.start(),
    ]
    assert fake_base_instance.mock_calls == [
        call.exists(),
        call.delete(),
        call.__bool__(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=False,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.config_set("user.craft_providers.timer", "2023-01-01T00:00:00+00:00"),
        call.lxc.config_set(
            instance_name="test-base-instance-e14661a426076717fa04",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.config_set("image.description", "test-base-instance-$"),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.stop(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.get_command_environment(),
        call.setup(executor=fake_base_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_all_opts(
    fake_instance,
    mock_base_configuration,
    mock_lxc,
    mock_lxd_instance,
    mock_timezone,
    mock_platform,
    mock_disable_timer_update_thread,
):
    """Parse all parameters."""
    fake_instance.config_get.return_value = "STARTING"
    fake_instance.lxc.config_get.return_value = "both 1234 0"

    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            auto_clean=True,
            auto_create_project=True,
            ephemeral=True,
            map_user_uid=True,
            uid=1234,
            project="test-project",
            remote="test-remote",
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("test-remote"),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=True,
            map_user_uid=True,
            uid=1234,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.lxc.config_set(
            instance_name=fake_instance.instance_name,
            key="raw.idmap",
            value="both 1234 0",
            project="test-project",
            remote="test-remote",
        ),
        call.lxc.config_get(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="raw.idmap",
            project="test-project",
            remote="test-remote",
        ),
        call.restart(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.setup(executor=fake_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_missing_project(
    fake_instance, mock_base_configuration, mock_lxc, mock_lxd_instance
):
    """Raise an error if project does not exist and auto_create_project if false."""
    with pytest.raises(lxd.LXDError) as exc_info:
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            auto_create_project=False,
            project="invalid-project",
            remote="test-remote",
            lxc=mock_lxc,
        )

    assert (
        exc_info.value.brief
        == "LXD project 'invalid-project' not found on remote 'test-remote'."
    )
    assert exc_info.value.details == "Available projects: ['default', 'test-project']"


def test_launch_create_project(
    fake_instance,
    mock_base_configuration,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
    mock_disable_timer_update_thread,
):
    """Create a project if it does not exist and auto_create_project is true."""
    fake_instance.config_get.return_value = "STARTING"
    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            auto_create_project=True,
            project="project-to-create",
            remote="test-remote",
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("test-remote"),
        call.project_create(project="project-to-create", remote="test-remote"),
        call.profile_show(profile="default", project="default", remote="test-remote"),
        call.profile_edit(
            profile="default",
            project="project-to-create",
            config=mock_lxc.profile_show.return_value,
            remote="test-remote",
        ),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="project-to-create",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=False,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="project-to-create",
            remote="test-remote",
        ),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.stop(),
        call.start(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.setup(executor=fake_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_with_existing_instance_not_running(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
):
    """If the existing instance is not running, start it."""
    fake_instance.exists.return_value = True

    lxd.launch(
        name=fake_instance.name,
        base_configuration=mock_base_configuration,
        image_name="image-name",
        image_remote="image-remote",
        lxc=mock_lxc,
    )

    assert mock_lxc.mock_calls == [call.project_list("local")]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="default",
            remote="local",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [call.exists(), call.is_running(), call.start()]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.warmup(executor=fake_instance),
    ]


def test_launch_with_existing_instance_running(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
):
    """If the existing instance is running, do not start it."""
    fake_instance.exists.return_value = True
    fake_instance.is_running.return_value = True

    lxd.launch(
        name=fake_instance.name,
        base_configuration=mock_base_configuration,
        image_name="image-name",
        image_remote="image-remote",
        lxc=mock_lxc,
    )

    assert mock_lxc.mock_calls == [call.project_list("local")]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="default",
            remote="local",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [call.exists(), call.is_running()]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.warmup(executor=fake_instance),
    ]


def test_launch_with_existing_instance_incompatible_with_auto_clean(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
    mock_disable_timer_update_thread,
):
    """If instance is incompatible and auto_clean is true, launch a new instance."""
    fake_instance.exists.return_value = True
    fake_instance.is_running.return_value = False
    fake_instance.config_get.return_value = "STARTING"

    mock_base_configuration.warmup.side_effect = [
        bases.BaseCompatibilityError(reason="foo"),
        None,
    ]

    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            auto_clean=True,
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("local"),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="default",
            remote="local",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.is_running(),
        call.start(),
        call.delete(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=False,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="default",
            remote="local",
        ),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.stop(),
        call.start(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.warmup(executor=fake_instance),
        call.setup(executor=fake_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_with_existing_instance_incompatible_without_auto_clean_error(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
):
    """If instance is incompatible and auto_clean is False, raise an error."""
    fake_instance.exists.return_value = True
    mock_base_configuration.warmup.side_effect = [
        bases.BaseCompatibilityError(reason="foo")
    ]

    with pytest.raises(bases.BaseCompatibilityError) as raised:
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            auto_clean=False,
            lxc=mock_lxc,
        )

    assert raised.value.brief == "Incompatible base detected: foo."
    assert raised.value.resolution == (
        "Clean incompatible instance and retry the requested operation."
    )


def test_launch_with_existing_ephemeral_instance(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
    mock_disable_timer_update_thread,
):
    """Delete and recreate existing ephemeral instances."""
    fake_instance.exists.return_value = True
    fake_instance.config_get.return_value = "STARTING"

    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            ephemeral=True,
            use_base_instance=False,
            lxc=mock_lxc,
        )

    assert mock_lxc.mock_calls == [
        call.project_list("local"),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="default",
            remote="local",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.delete(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=True,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="default",
            remote="local",
        ),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.restart(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.setup(executor=fake_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


def test_launch_existing_instance_id_map_mismatch_with_auto_clean(
    fake_instance,
    logs,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
):
    """If the id map is incorrect and auto_clean is true, return False."""
    fake_instance.exists.return_value = True
    mock_check_id_map.return_value = False

    result = lxd.launcher._launch_existing_instance(
        instance=fake_instance,
        lxc=mock_lxc,
        base_configuration=mock_base_configuration,
        project="test-project",
        remote="test-remote",
        auto_clean=True,
        ephemeral=False,
        map_user_uid=True,
        uid=1234,
    )

    assert not result

    assert (
        Exact(
            f"Cleaning incompatible instance '{fake_instance.instance_name}' (reason: "
            "the instance's id map ('raw.idmap') is not configured as expected)."
        )
        in logs.debug
    )


def test_launch_existing_instance_id_map_mismatch_without_auto_clean(
    fake_instance,
    mock_base_configuration,
    mock_check_id_map,
    mock_lxc,
    mock_lxd_instance,
):
    """If the id map is incorrect and auto_clean is False, raise an error."""
    fake_instance.exists.return_value = True
    mock_check_id_map.return_value = False

    with pytest.raises(bases.BaseCompatibilityError) as raised:
        lxd.launcher._launch_existing_instance(
            instance=fake_instance,
            lxc=mock_lxc,
            base_configuration=mock_base_configuration,
            project="test-project",
            remote="test-remote",
            auto_clean=False,
            ephemeral=False,
            map_user_uid=True,
            uid=1234,
        )

    assert raised.value.brief == (
        "Incompatible base detected: "
        "the instance's id map ('raw.idmap') is not configured as expected."
    )


def test_name_matches_base_name(
    fake_instance,
    fake_base_instance,
    mock_base_configuration,
    mock_lxc,
    mock_lxd_instance,
):
    """Raise an error if the instance name matches the base instance name."""
    # force the names to be equal
    fake_instance.instance_name = fake_base_instance.instance_name

    with pytest.raises(ProviderError) as raised:
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            use_base_instance=True,
            lxc=mock_lxc,
        )

    assert raised.value.brief == (
        f"instance name cannot match the base instance name: "
        f"{fake_base_instance.instance_name!r}"
    )
    assert raised.value.resolution == "change name of instance"


def test_use_snapshots_deprecated(
    fake_instance,
    fake_base_instance,
    logs,
    mock_is_valid,
    mock_base_configuration,
    mock_lxc,
    mock_lxd_instance,
    mock_platform,
    mock_timezone,
    mock_disable_timer_update_thread,
):
    """Log deprecation warning for `use_snapshots` and continue to launch."""
    fake_base_instance.config_get.return_value = "STARTING"

    with freeze_time("2023-01-01"):
        lxd.launch(
            name=fake_instance.name,
            base_configuration=mock_base_configuration,
            image_name="image-name",
            image_remote="image-remote",
            use_snapshots=True,
            project="test-project",
            remote="test-remote",
            lxc=mock_lxc,
        )

    assert (
        Exact(
            "Deprecated: Parameter 'use_snapshots' is deprecated. "
            "Use parameter 'use_base_instance' instead."
        )
        in logs.warning
    )

    assert mock_lxc.mock_calls == [
        call.project_list("test-remote"),
        call.copy(
            source_remote="test-remote",
            source_instance_name=fake_base_instance.instance_name,
            destination_remote="test-remote",
            destination_instance_name=fake_instance.instance_name,
            project="test-project",
        ),
    ]
    assert mock_lxd_instance.mock_calls == [
        call(
            name=fake_instance.name,
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
        call(
            name="base-instance-mock-compat-tag-v200-image-remote-image-name",
            project="test-project",
            remote="test-remote",
            default_command_environment={"foo": "bar"},
        ),
    ]
    assert fake_instance.mock_calls == [
        call.exists(),
        call.lxc.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.start(),
    ]
    assert fake_base_instance.mock_calls == [
        call.exists(),
        call.__bool__(),
        call.launch(
            image="image-name",
            image_remote="image-remote",
            ephemeral=False,
            map_user_uid=False,
            uid=None,
        ),
        call.config_get("user.craft_providers.status"),
        call.config_set("user.craft_providers.status", "PREPARING"),
        call.lxc.config_set(
            instance_name="test-base-instance-e14661a426076717fa04",
            key="environment.TZ",
            value="fake/timezone",
            project="test-project",
            remote="test-remote",
        ),
        call.config_set("image.description", "test-base-instance-$"),
        call.config_set("user.craft_providers.status", "FINISHED"),
        call.stop(),
    ]
    assert mock_base_configuration.mock_calls == [
        call.get_command_environment(),
        call.get_command_environment(),
        call.setup(executor=fake_base_instance),
        call.wait_until_ready(executor=fake_instance),
    ]


@freeze_time("2022/12/07 11:05:00 UTC")
@pytest.mark.parametrize(
    "creation_date",
    [
        "2022/09/08 11:05 UTC",  # 90 days old
        "2022/12/06 11:05 UTC",  # 1 day old
        "2022/12/08 11:05 UTC",  # 1 day in the future (improbable but valid)
    ],
)
def test_is_valid(creation_date, mocker, mock_lxc):
    """Instances younger than the expiration date (inclusive) are valid."""
    mock_lxc.info.return_value = {"Created": creation_date}

    is_valid = lxd.launcher._is_valid(
        instance_name="test-name",
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
        expiration=timedelta(days=90),
    )

    assert is_valid


@freeze_time("2022/12/07 11:05:00 UTC")
def test_is_valid_expired(mocker, mock_lxc):
    """Instances older than the expiration date are not valid."""
    # 91 days old
    mock_lxc.info.return_value = {"Created": "2022/09/07 11:05 UTC"}

    is_valid = lxd.launcher._is_valid(
        instance_name="test-name",
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
        expiration=timedelta(days=90),
    )

    assert not is_valid


def test_is_valid_lxd_error(logs, mocker, mock_lxc):
    """Warn if the instance's info cannot be retrieved."""
    mock_lxc.info.side_effect = lxd.LXDError("test error")

    is_valid = lxd.launcher._is_valid(
        instance_name="test-name",
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
        expiration=timedelta(days=1),
    )

    assert not is_valid
    assert Exact("Could not get instance info with error: test error") in logs.debug


def test_is_valid_key_error(logs, mocker, mock_lxc):
    """Warn if the instance does not have a creation date."""
    mock_lxc.info.return_value = {}

    is_valid = lxd.launcher._is_valid(
        instance_name="test-name",
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
        expiration=timedelta(days=1),
    )

    assert not is_valid
    assert Exact("Instance does not have a 'Created' date.") in logs.debug


def test_is_valid_value_error(logs, mocker, mock_lxc):
    """Warn if the instance's creation date cannot be parsed."""
    mock_lxc.info.return_value = {"Created": "bad-datetime-value"}

    is_valid = lxd.launcher._is_valid(
        instance_name="test-name",
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
        expiration=timedelta(days=1),
    )

    assert not is_valid
    assert (
        Exact(
            "Could not parse instance's 'Created' date with error: ValueError(\"time "
            "data 'bad-datetime-value' does not match format '%Y/%m/%d %H:%M %Z'\")"
        )
        in logs.debug
    )


@pytest.mark.skipif(sys.platform == "win32", reason="unsupported on windows")
def test_set_id_map_default(fake_base_instance, mock_lxc, mocker):
    """Verify `_set_id_map()` sets the id map with default arguments."""
    mocker.patch("craft_providers.lxd.launcher.os.getuid", return_value=101)
    mock_lxc.config_get.return_value = "both 101 0"

    lxd.launcher._set_id_map(instance=fake_base_instance, lxc=mock_lxc)

    assert mock_lxc.config_set.mock_calls == [
        call(
            instance_name=fake_base_instance.instance_name,
            key="raw.idmap",
            value="both 101 0",
            project="default",
            remote="local",
        )
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="unsupported on windows")
def test_set_id_map_default_fail(fake_base_instance, mock_lxc, mocker):
    """Verify `_set_id_map()` will fail if the id map is not usable."""
    mocker.patch("craft_providers.lxd.launcher.os.getuid", return_value=101)
    mocker.patch("time.sleep")
    mock_lxc.config_get.return_value = ""

    with pytest.raises(LXDError):
        lxd.launcher._set_id_map(instance=fake_base_instance, lxc=mock_lxc)


@pytest.mark.skipif(sys.platform == "win32", reason="unsupported on windows")
def test_set_id_map_all_options(fake_base_instance, mock_lxc, mocker):
    """Verify `_set_id_map()` sets the id map with all parameters specified."""
    mocker.patch("craft_providers.lxd.launcher.os.getuid", return_value=101)
    mock_lxc.config_get.return_value = "both 202 0"

    lxd.launcher._set_id_map(
        instance=fake_base_instance,
        lxc=mock_lxc,
        project="test-project",
        remote="test-remote",
        uid=202,
    )

    assert mock_lxc.config_set.mock_calls == [
        call(
            instance_name=fake_base_instance.instance_name,
            key="raw.idmap",
            value="both 202 0",
            project="test-project",
            remote="test-remote",
        )
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="unsupported on windows")
@pytest.mark.parametrize(
    ("map_user_uid", "actual_uid", "expected_uid", "expected_result"),
    [
        # return True if an id map is not expected and there is no id map
        (False, None, None, True),
        # return False if an id map is expected and there is no id map
        (True, None, 5678, False),
        # return False if id map is not expected and there is an id map
        (False, 1234, None, False),
        # return True if an id map is expected and the uid matches
        (True, 1234, 1234, True),
        # return False if an id map is expected and the uid does not match
        (True, 1234, 5678, False),
        # return True if an id map is expected and the uid matches the current users uid
        (True, 101, None, True),
    ],
)
def test_check_id_map(
    map_user_uid,
    expected_uid,
    actual_uid,
    expected_result,
    fake_base_instance,
    mock_lxc,
    mocker,
):
    """Verify the instances id map is properly checked."""
    mocker.patch("craft_providers.lxd.launcher.os.getuid", return_value=101)

    if actual_uid:
        mock_lxc.config_get.return_value = f"both {actual_uid} 0"
    else:
        mock_lxc.config_get.return_value = ""

    result = lxd.launcher._check_id_map(
        instance=fake_base_instance,
        lxc=mock_lxc,
        project="test-project",
        remote="test-remote",
        map_user_uid=map_user_uid,
        uid=expected_uid,
    )

    assert result == expected_result


def test_check_id_map_wrong_format(fake_base_instance, logs, mock_lxc, mocker):
    """Return false if the id map is not formatted as expected."""
    mock_lxc.config_get.return_value = "gid 100-200 300-400"

    result = lxd.launcher._check_id_map(
        instance=fake_base_instance,
        lxc=mock_lxc,
        project="test-project",
        remote="test-remote",
        map_user_uid=True,
        uid=1234,
    )

    assert not result
    assert (
        Exact(
            f"Unexpected id map for '{fake_base_instance.instance_name}' "
            "(expected 'both 1234 0', got 'gid 100-200 300-400')."
        )
        in logs.debug
    )


@pytest.mark.parametrize("platform", ["darwin", "win32", "other"])
def test_timezone_non_linux_host(fake_instance, mocker, mock_lxc, logs, platform):
    """Log an error and no-op if host is not linux."""
    mocker.patch.object(sys, "platform", platform)

    lxd.launcher._set_timezone(
        instance=fake_instance,
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
    )

    assert "Not setting timezone because host is not Linux." in logs.debug


def test_timezone_host_error(
    fake_instance, fake_process, mock_lxc, mock_platform, logs
):
    """Log an error and no-op if timezone cannot be collected from host."""
    fake_process.register_subprocess(
        ["timedatectl", "show", "-p", "Timezone", "--value"],
        returncode=1,
    )

    lxd.launcher._set_timezone(
        instance=fake_instance,
        project="test-project",
        remote="test-remote",
        lxc=mock_lxc,
    )

    assert (
        "Not setting instance's timezone because host timezone could not "
        "be determined: \\* Command that failed: 'timedatectl show -p Timezone --value'"
    ) in logs.debug


def test_timer_error_ignore(fake_instance, fake_process, mock_lxc, mocker):
    """LXC timer should ignore errors."""
    mocker.patch("time.sleep")

    fake_instance.config_set.side_effect = LXDError("test error")
    timer = lxd.launcher.InstanceTimer(fake_instance)
    timer.start()
    timer.stop()

    assert fake_instance.config_set.call_count > 0



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass Instance."""
import io
import logging
import pathlib
import subprocess
from typing import Any, Dict, List, Optional

from craft_providers import errors
from craft_providers.const import TIMEOUT_COMPLEX, TIMEOUT_SIMPLE
from craft_providers.util import env_cmd

from .. import Executor
from .errors import MultipassError
from .multipass import Multipass

logger = logging.getLogger(__name__)


def _rootify_multipass_command(
    command: List[str],
    *,
    cwd: Optional[pathlib.PurePath] = None,
    env: Optional[Dict[str, Optional[str]]] = None,
) -> List[str]:
    """Wrap a command to run as root with specified environment.

    - Use sudo to run as root (Multipass defaults to ubuntu user).
    - Configure sudo to set home directory.
    - Account for environment flags in env, if any.

    :param command: Command to execute.
    :param env: Additional environment flags to set.

    :returns: List of command strings for multipass exec.
    """
    sudo_cmd = ["sudo", "-H", "--"]

    if env is not None or cwd is not None:
        sudo_cmd += env_cmd.formulate_command(env, chdir=cwd)

    return sudo_cmd + command


class MultipassInstance(Executor):
    """Multipass Instance Lifecycle.

    :param name: Name of multipass instance.
    """

    def __init__(
        self,
        *,
        name: str,
        multipass: Optional[Multipass] = None,
    ) -> None:
        super().__init__()

        self.name = name

        if multipass is not None:
            self._multipass = multipass
        else:
            self._multipass = Multipass()

    def _create_temp_file(self) -> str:
        """Create a temporary file inside the instance owned by the `ubuntu` user.

        :returns: String containing path to temporary file.

        :raises subprocess.CalledProcessError: If the file cannot be created.
        """
        tmp_file_path = self.execute_run(
            command=["mktemp"],
            capture_output=True,
            check=True,
            text=True,
            timeout=TIMEOUT_SIMPLE,
        ).stdout.strip()

        # mktemp is executed as root, so the ownership of the temp file needs to be
        # changed back to the default user `ubuntu` before transferring the file
        self.execute_run(
            ["chown", "ubuntu:ubuntu", tmp_file_path],
            capture_output=True,
            check=True,
            timeout=TIMEOUT_SIMPLE,
        )

        logger.debug("Created temporary file %r inside instance.", tmp_file_path)

        return tmp_file_path

    def _is_dir_in_instance(self, filepath: pathlib.PurePath) -> bool:
        """Check if a filepath inside a Multipass instance is a valid directory.

        :param filepath: filepath to check

        :returns: True if the filepath is a valid directory.
        """
        proc = self.execute_run(
            ["test", "-d", filepath.as_posix()], timeout=TIMEOUT_SIMPLE, check=False
        )
        return proc.returncode == 0

    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        """Create or replace file with content and file mode.

        Multipass transfers data as "ubuntu" user, forcing us to first copy a
        file to a temporary location before moving to a (possibly) root-owned
        location and with appropriate permissions.

        :param destination: Path to file.
        :param content: Contents of file.
        :param file_mode: File mode string (e.g. '0644').
        :param group: File group owner/id.
        :param user: File user owner/id.

        :raises MultipassError: If the content cannot be pushed into the instance.
        """
        try:
            tmp_file_path = self._create_temp_file()

            self._multipass.transfer_source_io(
                source=content, destination=f"{self.name}:{tmp_file_path}"
            )

            # now that the file has been transferred, its ownership can be set
            self.execute_run(
                ["chown", f"{user}:{group}", tmp_file_path],
                capture_output=True,
                check=True,
                timeout=TIMEOUT_SIMPLE,
            )

            self.execute_run(
                ["chmod", file_mode, tmp_file_path],
                capture_output=True,
                check=True,
                timeout=TIMEOUT_SIMPLE,
            )

            self.execute_run(
                ["mv", tmp_file_path, destination.as_posix()],
                capture_output=True,
                check=True,
                timeout=TIMEOUT_COMPLEX,
            )
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=(
                    f"Failed to create file {destination.as_posix()!r}"
                    f" in Multipass instance {self.name!r}."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def delete(self) -> None:
        """Delete instance and purge."""
        return self._multipass.delete(
            instance_name=self.name,
            purge=True,
        )

    def execute_popen(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> subprocess.Popen:
        """Execute a process in the instance using subprocess.Popen().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        The command is run as root via sudo. Running as root may be required even
        when the command itself does not require root permissions, because the
        instance's working directory may be a directory that the default `ubuntu` user
        does not have access to.

        :param command: Command to execute.
        :param cwd: working directory to execute the command
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param kwargs: Additional keyword arguments for subprocess.Popen().

        :returns: Popen instance.
        """
        return self._multipass.exec(
            instance_name=self.name,
            command=_rootify_multipass_command(command, cwd=cwd, env=env),
            runner=subprocess.Popen,
            timeout=timeout,
            **kwargs,
        )

    def execute_run(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        check: bool = False,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Execute a command in the instance using subprocess.run().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        The command is run as root via sudo. Running as root may be required even
        when the command itself does not require root permissions, because the
        instance's working directory may be a directory that the default `ubuntu` user
        does not have access to.

        :param command: Command to execute.
        :param cwd: working directory to execute the command
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param check: Raise an exception if the command fails.
        :param kwargs: Keyword args to pass to subprocess.run().

        :returns: Completed process.

        :raises subprocess.CalledProcessError: if command fails and check is True.
        """
        return self._multipass.exec(
            instance_name=self.name,
            command=_rootify_multipass_command(command, cwd=cwd, env=env),
            runner=subprocess.run,
            timeout=timeout,
            check=check,
            **kwargs,
        )

    def exists(self) -> bool:
        """Check if instance exists.

        :returns: True if instance exists.

        :raises MultipassError: On unexpected failure.
        """
        vm_list = self._multipass.list()

        return self.name in vm_list

    def _get_info(self) -> Dict[str, Any]:
        """Get configuration and state for instance.

        :returns: State information parsed from multipass if instance exists,
            else None.

        :raises MultipassError: If unable to parse VM info.
        """
        info_data = self._multipass.info(instance_name=self.name).get("info")

        if info_data is None or self.name not in info_data:
            raise MultipassError(
                brief="Malformed multipass info",
                details=f"Returned data: {info_data!r}",
            )

        return info_data[self.name]

    def is_mounted(
        self, *, host_source: pathlib.Path, target: pathlib.PurePath
    ) -> bool:
        """Check if path is mounted at target.

        :param host_source: Host path to check.
        :param target: Instance path to check.

        :returns: True if host_source is mounted at target.

        :raises MultipassError: On unexpected failure.
        """
        info = self._get_info()
        mounts = info.get("mounts", {})

        for mount_point, mount_config in mounts.items():
            # Even on Windows, Multipass writes source_path as posix, e.g.:
            # 'C:/Users/chris/tmpbat91bwz.tmp-pytest' # noqa: ERA001
            if (
                mount_point == target.as_posix()
                and mount_config.get("source_path") == host_source.as_posix()
            ):
                return True

        return False

    def is_running(self) -> bool:
        """Check if instance is running.

        :returns: True if instance is running.

        :raises MultipassError: On unexpected failure.
        """
        info = self._get_info()

        return info.get("state") == "Running"

    def launch(
        self,
        *,
        image: str,
        cpus: int = 2,
        disk_gb: int = 256,
        mem_gb: int = 2,
    ) -> None:
        """Launch instance.

        :param image: Name of image to create the instance with.
        :param instance_cpus: Number of CPUs.
        :param instance_disk_gb: Disk allocation in gigabytes.
        :param instance_mem_gb: Memory allocation in gigabytes.
        :param instance_name: Name of instance to use/create.
        :param instance_stop_time_mins: Stop time delay in minutes.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.launch(
            instance_name=self.name,
            image=image,
            cpus=str(cpus),
            disk=f"{disk_gb!s}G",
            mem=f"{mem_gb!s}G",
        )

    def mount(
        self,
        *,
        host_source: pathlib.Path,
        target: pathlib.PurePath,
    ) -> None:
        """Mount host host_source directory to target mount point.

        Checks first to see if already mounted.

        :param host_source: Host path to mount.
        :param target: Instance path to mount to.

        :raises MultipassError: On unexpected failure.
        """
        if self.is_mounted(host_source=host_source, target=target):
            return

        self._multipass.mount(
            source=host_source,
            target=f"{self.name}:{target.as_posix()}",
        )

    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        """Copy a file from the environment to host.

        :param source: Environment file to copy.
        :param destination: Host file path to copy to.  Parent directory
            (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises MultipassError: On unexpected error copying file.
        """
        proc = self.execute_run(
            ["test", "-f", source.as_posix()],
            check=False,
            timeout=TIMEOUT_SIMPLE,
        )
        if proc.returncode != 0:
            raise FileNotFoundError(f"File not found: {source.as_posix()!r}")

        if not destination.parent.is_dir():
            raise FileNotFoundError(f"Directory not found: {str(destination.parent)!r}")

        self._multipass.transfer(
            source=f"{self.name}:{source.as_posix()}", destination=str(destination)
        )

    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        """Copy a file from the host into the environment.

        The destination file is overwritten if it exists. File permissions are retained
        but the ownership is changed to the default user `ubuntu`.

        The source cannot be a directory because `multipass transfer --recursive` is not
        supported. The parent directories of the destination must exist because
        `multipass transfer --parents` is not supported.

        :param source: Host file to copy.
        :param destination: Target environment file path to copy to.  Parent
            directory (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist or if the source is not a regular file.
        :raises IsADirectoryError: If source is a directory.
        :raises MultipassError: If the file cannot be pushed into the instance.
        """
        if source.is_dir():
            raise IsADirectoryError(f"Source cannot be a directory: {str(source)!r}")

        if not source.is_file():
            raise FileNotFoundError(f"File not found: {str(source)!r}")

        if not self._is_dir_in_instance(destination.parent):
            raise FileNotFoundError(
                "Directory not found in instance: "
                f"{str(destination.parent.as_posix())!r}"
            )

        try:
            tmp_file_path = self._create_temp_file()

            # push the file to a location where the `ubuntu` user has access
            self._multipass.transfer(
                source=str(source), destination=f"{self.name}:{tmp_file_path}"
            )

            # if the destination was a directory, then we need to specify the filename
            # to prevent using the random name generated by mktemp
            if self._is_dir_in_instance(destination):
                final_destination = destination / source.name
            else:
                final_destination = destination

            # move the file to its final destination
            self.execute_run(
                ["mv", tmp_file_path, final_destination.as_posix()],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=(
                    f"Failed to push file {destination.as_posix()!r}"
                    f" into Multipass instance {self.name!r}."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def start(self) -> None:
        """Start instance.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.start(instance_name=self.name)

    def stop(self, *, delay_mins: int = 0) -> None:
        """Stop instance.

        :param delay_mins: Delay shutdown for specified minutes.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.stop(instance_name=self.name, delay_mins=delay_mins)

    def unmount(self, target: pathlib.Path) -> None:
        """Unmount mount target shared with host.

        :param target: Target shared with host to unmount.

        :raises MultipassError: On failure to unmount target.
        """
        mount = f"{self.name}:{target.as_posix()}"

        self._multipass.umount(mount=mount)

    def unmount_all(self) -> None:
        """Unmount all mounts shared with host.

        :raises MultipassError: On failure to unmount target.
        """
        self._multipass.umount(mount=self.name)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import pathlib
import subprocess
import sys
from pathlib import Path
from textwrap import dedent
from unittest.mock import ANY, call, patch

import pytest
from craft_providers.actions.snap_installer import Snap, SnapInstallationError
from craft_providers.bases import ubuntu
from craft_providers.errors import (
    BaseCompatibilityError,
    BaseConfigurationError,
    NetworkError,
    details_from_called_process_error,
)
from craft_providers.instance_config import InstanceConfiguration
from logassert import Exact  # type: ignore
from pydantic import ValidationError

from tests.unit.conftest import DEFAULT_FAKE_CMD


@pytest.fixture()
def mock_load(mocker):
    return mocker.patch(
        "craft_providers.instance_config.InstanceConfiguration.load",
        return_value=InstanceConfiguration(compatibility_tag="buildd-base-v2"),
    )


@pytest.fixture()
def fake_filesystem(fs):
    return fs


@pytest.fixture()
def mock_install_from_store(mocker):
    return mocker.patch("craft_providers.actions.snap_installer.install_from_store")


@pytest.fixture()
def mock_inject_from_host(mocker):
    return mocker.patch("craft_providers.actions.snap_installer.inject_from_host")


@pytest.fixture()
def mock_get_os_release(mocker):
    return mocker.patch.object(
        ubuntu.BuilddBase,
        "_get_os_release",
        return_value={
            "NAME": "Ubuntu",
            "VERSION_ID": "22.04",
            "VERSION_CODENAME": "jammy",
        },
    )


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize(
    ("environment", "etc_environment_content"),
    [
        (
            None,
            (
                b"PATH=/usr/local/sbin:/usr/local/bin:"
                b"/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin\n"
            ),
        ),
        (
            {
                "https_proxy": "http://foo.bar:8081",
                "PATH": "/snap",
                "http_proxy": "http://foo.bar:8080",
            },
            (
                b"https_proxy=http://foo.bar:8081\n"
                b"PATH=/snap\nhttp_proxy=http://foo.bar:8080\n"
            ),
        ),
    ],
)
@pytest.mark.parametrize("no_cdn", [False, True])
@pytest.mark.parametrize(
    ("snaps", "expected_snap_call"),
    [
        (None, []),
        (
            [Snap(name="snap1", channel="edge", classic=True)],
            [call(executor=ANY, snap_name="snap1", channel="edge", classic=True)],
        ),
    ],
)
@pytest.mark.parametrize(
    ("packages", "expected_packages"),
    [
        (
            None,
            [
                "apt-utils",
                "build-essential",
                "curl",
                "fuse",
                "udev",
                "python3",
                "python3-dev",
                "python3-pip",
                "python3-wheel",
                "python3-setuptools",
            ],
        ),
        (
            ["grep", "git"],
            [
                "apt-utils",
                "build-essential",
                "curl",
                "fuse",
                "udev",
                "python3",
                "python3-dev",
                "python3-pip",
                "python3-wheel",
                "python3-setuptools",
                "grep",
                "git",
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("tag", "expected_tag"), [(None, "buildd-base-v2"), ("test-tag", "test-tag")]
)
def test_setup(
    fake_process,
    fake_executor,
    fake_filesystem,
    alias,
    environment,
    etc_environment_content,
    no_cdn,
    mock_load,
    mock_inject_from_host,
    mock_install_from_store,
    mocker,
    snaps,
    expected_snap_call,
    packages,
    expected_packages,
    tag,
    expected_tag,
):
    mock_load.return_value = InstanceConfiguration(compatibility_tag=expected_tag)

    if environment is None:
        environment = ubuntu.BuilddBase.default_command_environment()

    if no_cdn:
        fake_filesystem.create_file(
            "/etc/systemd/system/snapd.service.d/no-cdn.conf",
            contents=dedent(
                """\
                [Service]
                Environment=SNAPPY_STORE_NO_CDN=1
                """
            ),
        )

    base_config = ubuntu.BuilddBase(
        alias=alias,
        compatibility_tag=tag,
        environment=environment,
        hostname="test-hostname",
        snaps=snaps,
        packages=packages,
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", "-f", "/etc/craft-instance.conf"],
        returncode=1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "hostname", "-F", "/etc/hostname"]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-resolved"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-networkd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            VERSION_CODENAME="test-name"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "sed", "-i", "s/test-name/devel/g", "/etc/apt/sources.list"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", "-s", "/etc/cloud/cloud.cfg"]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "$ aapt_preserve_sources_list: true",
            "/etc/cloud/cloud.cfg",
        ]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "find",
            "/etc/apt/sources.list.d/",
            "-type",
            "f",
            "-name",
            "*.list",
        ]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", *expected_packages]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "fuse", "udev"]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "autoremove", "-y"])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "clean", "-y"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-active", "systemd-udevd"],
        stdout="active",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-udevd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "systemd-udevd"]
    )
    if no_cdn:
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/etc/systemd/system/snapd.service.d"]
        )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "snapd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "snapd.socket"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "snapd.service"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "wait", "system", "seed.loaded"]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.http=http://foo.bar:8080"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.https=http://foo.bar:8081"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "ln", "-sf", "/var/lib/snapd/snap", "/snap"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "--now", "snapd.socket"]
    )

    base_config.setup(executor=fake_executor)

    expected_push_file_io = [
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\nsetup: false\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/environment",
            "content": etc_environment_content,
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/20auto-upgrades",
            "content": dedent(
                """\
                APT::Periodic::Update-Package-Lists "10000";
                APT::Periodic::Unattended-Upgrade "0";
                """
            ).encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/hostname",
            "content": b"test-hostname\n",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/systemd/network/10-eth0.network",
            "content": dedent(
                """\
                [Match]
                Name=eth0

                [Network]
                DHCP=ipv4
                LinkLocalAddressing=ipv6

                [DHCP]
                RouteMetric=100
                UseMTU=true
                """
            ).encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/00no-recommends",
            "content": b'APT::Install-Recommends "false";\n',
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/00update-errors",
            "content": b'APT::Update::Error-Mode "any";\n',
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\nsetup: true\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
    ]
    expected_push_file = []
    if no_cdn:
        expected_push_file.append(
            {
                "source": Path("/etc/systemd/system/snapd.service.d/no-cdn.conf"),
                "destination": Path("/etc/systemd/system/snapd.service.d/no-cdn.conf"),
            }
        )

    assert fake_executor.records_of_push_file_io == expected_push_file_io
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == expected_push_file
    assert mock_install_from_store.mock_calls == expected_snap_call


def test_install_snaps_install_from_store(fake_executor, mock_install_from_store):
    """Verify installing snaps calls install_from_store()."""
    my_snaps = [
        Snap(name="snap1"),
        Snap(name="snap2", channel="edge"),
        Snap(name="snap3", channel="edge", classic=True),
    ]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    base._install_snaps(executor=fake_executor)

    assert mock_install_from_store.mock_calls == [
        call(
            executor=fake_executor, snap_name="snap1", channel="stable", classic=False
        ),
        call(executor=fake_executor, snap_name="snap2", channel="edge", classic=False),
        call(executor=fake_executor, snap_name="snap3", channel="edge", classic=True),
    ]


def test_install_snaps_inject_from_host_valid(
    fake_executor, mock_inject_from_host, mocker
):
    """Verify installing snaps calls inject_from_host()."""
    mocker.patch("sys.platform", "linux")
    my_snaps = [
        Snap(name="snap1", channel=None),
        Snap(name="snap2", channel=None, classic=True),
    ]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    base._install_snaps(executor=fake_executor)

    assert mock_inject_from_host.mock_calls == [
        call(executor=fake_executor, snap_name="snap1", classic=False),
        call(executor=fake_executor, snap_name="snap2", classic=True),
    ]


def test_install_snaps_inject_from_host_not_linux_error(fake_executor, mocker):
    """Verify install_snaps raises an error when injecting from host on
    a non-linux system."""
    mocker.patch("sys.platform", return_value="darwin")
    my_snaps = [Snap(name="snap1", channel=None)]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="cannot inject snap 'snap1' from host on a non-linux system",
        resolution="install the snap from the store by setting the 'channel' parameter",
    )


def test_install_snaps_install_from_store_error(fake_executor, mocker):
    """Verify install_snaps raises an error when install_from_store fails."""
    mocker.patch(
        "craft_providers.actions.snap_installer.install_from_store",
        side_effect=SnapInstallationError(brief="test error"),
    )
    my_snaps = [Snap(name="snap1", channel="candidate")]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief=(
            "failed to install snap 'snap1' from store"
            " channel 'candidate' in target environment."
        )
    )


def test_install_snaps_inject_from_host_error(fake_executor, mocker):
    """Verify install_snaps raises an error when inject_from_host fails."""
    mocker.patch("sys.platform", "linux")
    mocker.patch(
        "craft_providers.actions.snap_installer.inject_from_host",
        side_effect=SnapInstallationError(brief="test error"),
    )
    my_snaps = [Snap(name="snap1", channel=None)]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="failed to inject host's snap 'snap1' into target environment."
    )


def test_setup_apt(fake_executor, fake_process):
    """Verify packages are installed as expected."""
    packages = ["grep", "git"]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, packages=packages)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "apt-utils",
            "build-essential",
            "curl",
            "fuse",
            "udev",
            "python3",
            "python3-dev",
            "python3-pip",
            "python3-wheel",
            "python3-setuptools",
            "grep",
            "git",
        ]
    )

    base._setup_packages(executor=fake_executor)


def test_setup_apt_install_default(fake_executor, fake_process):
    """Verify only default packages are installed."""
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "apt-utils",
            "build-essential",
            "curl",
            "fuse",
            "udev",
            "python3",
            "python3-dev",
            "python3-pip",
            "python3-wheel",
            "python3-setuptools",
        ]
    )

    base._setup_packages(executor=fake_executor)


def test_setup_apt_install_override_system(fake_executor, fake_process):
    """Verify override default packages."""
    base = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        packages=["clang"],
        use_default_packages=False,
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "clang",
        ]
    )

    base._setup_packages(executor=fake_executor)


def test_setup_apt_install_packages_update_error(mocker, fake_executor):
    """Verify error is caught from `apt-get update` call."""
    error = subprocess.CalledProcessError(100, ["error"])
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    mocker.patch.object(fake_executor, "execute_run", side_effect=error)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._pre_setup_packages(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to update apt cache.",
        details="* Command that failed: 'error'\n* Command exit code: 100",
    )


def test_setup_apt_install_packages_install_error(mocker, fake_executor):
    """Verify error is caught from `apt-get install` call."""
    error = subprocess.CalledProcessError(100, ["error"])
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    side_effects = [
        error,  # make apt-get install fail
        subprocess.CompletedProcess("args", returncode=0),  # network connectivity check
    ]
    mocker.patch.object(fake_executor, "execute_run", side_effect=side_effects)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._setup_packages(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to install packages.",
        details="* Command that failed: 'error'\n* Command exit code: 100",
    )


def test_pre_setup_packages_devel(fake_executor, fake_process, mocker):
    """Verify `update_apt_sources()` is called for devel bases."""
    mock_update_apt_sources = mocker.patch.object(
        ubuntu.BuilddBase, "_update_apt_sources"
    )

    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.DEVEL)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "apt-utils",
            "build-essential",
            "python3",
            "python3-dev",
            "python3-pip",
            "python3-wheel",
            "python3-setuptools",
            "curl",
            "fuse",
            "udev",
        ]
    )

    base._pre_setup_packages(executor=fake_executor)

    mock_update_apt_sources.assert_called_once()


def test_ensure_image_version_compatible_failure(fake_executor, monkeypatch):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    monkeypatch.setattr(
        InstanceConfiguration,
        "load",
        lambda **kwargs: InstanceConfiguration(compatibility_tag="invalid-tag"),
    )

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_instance_config_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected image compatibility tag 'buildd-base-v2', found 'invalid-tag'"
    )


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize("cache_path", [pathlib.Path("/tmp/fake-cache-dir")])
def test_mount_cache_dirs(fake_process, fake_executor, cache_path, alias):
    """Test mounting of cache directories with a cache directory set."""
    base = ubuntu.BuilddBase(alias=alias, cache_path=cache_path)
    host_cache_dir = cache_path / base.compatibility_tag / str(base.alias)
    user_cache_dir = pathlib.Path("/root/.cache")
    fake_process.register(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "echo -n ${XDG_CACHE_HOME:-${HOME}/.cache}"],
        stdout=str(user_cache_dir),
    )
    fake_process.register(
        [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/root/.cache/pip"],
    )

    base._mount_shared_cache_dirs(fake_executor)

    if sys.platform == "win32":
        expected_mounts = [
            {
                "host_source": pathlib.WindowsPath("D:") / host_cache_dir / "pip",
                "target": user_cache_dir / "pip",
            }
        ]
    else:
        expected_mounts = [
            {
                "host_source": host_cache_dir / "pip",
                "target": user_cache_dir / "pip",
            },
        ]
    assert fake_executor.records_of_mount == expected_mounts


def test_get_os_release(fake_process, fake_executor):
    """`_get_os_release` should parse data from `/etc/os-release` to a dict."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="NAME=Ubuntu\nVERSION_ID=12.04\n",
    )

    result = base_config._get_os_release(executor=fake_executor)

    assert result == {"NAME": "Ubuntu", "VERSION_ID": "12.04"}


def test_ensure_os_compatible(fake_executor, fake_process, mock_get_os_release):
    """Do nothing if the OS is compatible."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    base_config._ensure_os_compatible(executor=fake_executor)

    mock_get_os_release.assert_called_once()


def test_ensure_os_compatible_devel_mismatch(
    fake_executor, fake_process, logs, mock_get_os_release
):
    """Ignore OS version id mismatch when using a devel base."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.DEVEL)

    base_config._ensure_os_compatible(executor=fake_executor)

    mock_get_os_release.assert_called_once()

    assert (
        "Ignoring OS version mismatch for '22.04' because base is 'devel'."
        in logs.debug
    )


def test_ensure_os_compatible_name_failure(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error if the OS name does not match."""
    mock_get_os_release.return_value = {"NAME": "Fedora", "VERSION_ID": "32"}
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_os_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected OS 'Ubuntu', found 'Fedora'"
    )

    mock_get_os_release.assert_called_once()


def test_ensure_os_compatible_version_failure(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error if the OS version id does not match."""
    mock_get_os_release.return_value = {"NAME": "Ubuntu", "VERSION_ID": "12.04"}
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_os_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected OS version '22.04', found '12.04'"
    )

    mock_get_os_release.assert_called_once()


def test_setup_hostname_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "hostname", "-F", "/etc/hostname"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_hostname(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to set hostname.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_setup_networkd_enable_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_networkd(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to setup systemd-networkd.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_setup_networkd_restart_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-networkd"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_networkd(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to setup systemd-networkd.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_setup_resolved_enable_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_resolved(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to setup systemd-resolved.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_setup_resolved_restart_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-resolved"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_resolved(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to setup systemd-resolved.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_setup_snapd_proxy(fake_executor, fake_process):
    """Verify snapd proxy is set or unset."""
    environment = {
        "http_proxy": "http://foo.bar:8080",
        "https_proxy": "http://foo.bar:8081",
    }
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        environment=environment,  # type: ignore
    )
    fake_process.keep_last_process(True)
    fake_process.register([fake_process.any()])

    base_config._setup_snapd_proxy(executor=fake_executor)
    assert [
        *DEFAULT_FAKE_CMD,
        "snap",
        "set",
        "system",
        "proxy.http=http://foo.bar:8080",
    ] in fake_process.calls
    assert [
        *DEFAULT_FAKE_CMD,
        "snap",
        "set",
        "system",
        "proxy.https=http://foo.bar:8081",
    ] in fake_process.calls


@pytest.mark.parametrize("fail_index", list(range(0, 1)))
def test_setup_snapd_proxy_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    return_codes = [0, 0]
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[1],
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_snapd_proxy(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to set the snapd proxy.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.usefixtures("stub_verify_network")
@pytest.mark.parametrize("fail_index", list(range(0, 2)))
def test_pre_setup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    return_codes = [0, 0]
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-active", "systemd-udevd"],
        stdout="inactive",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-udevd"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "systemd-udevd"],
        returncode=return_codes[1],
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._pre_setup_snapd(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to enable systemd-udevd service.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.usefixtures("stub_verify_network")
def test_setup_snapd_failures(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "snapd"],
        returncode=1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._setup_snapd(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to setup snapd.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.parametrize("fail_index", list(range(0, 8)))
def test_post_setup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    return_codes = [0] * 8
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "ln", "-sf", "/var/lib/snapd/snap", "/snap"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "--now", "snapd.socket"],
        returncode=return_codes[1],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "snapd.service"],
        returncode=return_codes[2],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "wait", "system", "seed.loaded"],
        returncode=return_codes[3],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"],
        returncode=return_codes[4],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"],
        returncode=return_codes[5],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[6],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[7],
    )

    with pytest.raises(BaseConfigurationError):
        base_config._post_setup_snapd(executor=fake_executor)


@pytest.mark.parametrize("fail_index", list(range(0, 2)))
def test_warmup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    return_codes = [0] * 2
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[1],
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._warmup_snapd(executor=fake_executor)

    assert raised.value == BaseConfigurationError(
        brief="Failed to set the snapd proxy.",
        details=details_from_called_process_error(
            raised.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize("system_running_ready_stdout", ["degraded", "running"])
def test_wait_for_system_ready(
    fake_executor, fake_process, alias, system_running_ready_stdout
):
    base_config = ubuntu.BuilddBase(alias=alias)
    base_config._retry_wait = 0.01
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout="not-ready",
        returncode=-1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="still-not-ready"
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout=system_running_ready_stdout,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=-1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=0,
    )

    base_config.wait_until_ready(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == []
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == []
    assert list(fake_process.calls) == [
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "getent",
            "hosts",
            "snapcraft.io",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "getent",
            "hosts",
            "snapcraft.io",
        ],
    ]


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
def test_wait_for_system_ready_timeout(fake_executor, fake_process, alias):
    base_config = ubuntu.BuilddBase(
        alias=alias,
    )
    base_config._timeout_simple = 0.01
    base_config._retry_wait = 0.01
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout="not-ready",
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config.wait_until_ready(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Timed out waiting for environment to be ready."
    )


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
def test_wait_for_system_ready_timeout_in_network(
    fake_executor, fake_process, alias, monkeypatch
):
    base_config = ubuntu.BuilddBase(alias=alias)
    base_config._timeout_simple = 0.01
    base_config._retry_wait = 0.01
    monkeypatch.setattr(
        base_config, "_setup_wait_for_system_ready", lambda **kwargs: None
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config.wait_until_ready(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Timed out waiting for networking to be ready."
    )


def test_update_apt_sources(fake_executor, fake_process, mock_get_os_release, logs):
    """`update_apt_sources()` should update the apt source config files."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "s/jammy/test-codename/g",
            "/etc/apt/sources.list",
        ]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", "-s", "/etc/cloud/cloud.cfg"]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "$ aapt_preserve_sources_list: true",
            "/etc/cloud/cloud.cfg",
        ]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "find",
            "/etc/apt/sources.list.d/",
            "-type",
            "f",
            "-name",
            "*.list",
        ],
    )

    base_config._update_apt_sources(executor=fake_executor, codename="test-codename")

    mock_get_os_release.assert_called_once()
    assert Exact("Updating apt sources from 'jammy' to 'test-codename'.") in logs.debug
    assert (
        Exact("Updating '/etc/cloud/cloud.cfg' to preserve apt sources.") in logs.debug
    )


def test_update_apt_sources_dir(fake_executor, fake_process, mock_get_os_release):
    """Verify source files in `/etc/apt/sources.list.d/` are updated."""
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "s/jammy/test-codename/g",
            "/etc/apt/sources.list",
        ]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", "-s", "/etc/cloud/cloud.cfg"]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "$ aapt_preserve_sources_list: true",
            "/etc/cloud/cloud.cfg",
        ]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "find",
            "/etc/apt/sources.list.d/",
            "-type",
            "f",
            "-name",
            "*.list",
        ],
        stdout="/etc/apt/sources.list.d/file1.list\n/etc/apt/sources.list.d/file2.list",
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "sed",
            "-i",
            "s/jammy/test-codename/g",
            "/etc/apt/sources.list.d/*.list",
        ]
    )

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    base_config._update_apt_sources(executor=fake_executor, codename="test-codename")

    mock_get_os_release.assert_called_once()


def test_update_apt_sources_source_list_sed_error(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error when the sed command fails to update apt sources."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    # fail on the first `sed` call
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "sed", fake_process.any()], returncode=1
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_apt_sources(
            executor=fake_executor, codename="test-codename"
        )

    assert raised.value.brief == "Failed to update '/etc/apt/sources.list'."


def test_update_apt_sources_cloud_cfg_does_not_exist_error(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error when cloud.cfg is empty or does not exist."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    # fail on the `test` call
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", fake_process.any()], returncode=1
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_apt_sources(
            executor=fake_executor, codename="test-codename"
        )

    assert raised.value.brief == (
        "Could not update '/etc/cloud/cloud.cfg' because it is empty or does not exist."
    )


def test_update_apt_sources_cloud_cfg_sed_error(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error when the sed command fails to update cloud.cfg."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    # fail on the second `sed` call
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "test", fake_process.any()])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "sed", fake_process.any()], returncode=1
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_apt_sources(
            executor=fake_executor, codename="test-codename"
        )

    assert raised.value.brief == "Failed to update '/etc/cloud/cloud.cfg'."


def test_update_apt_sources_find_error(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error when the find command fails to find apt source files."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "test", fake_process.any()])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    # fail on the `find` call
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "find", fake_process.any()], returncode=1
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_apt_sources(
            executor=fake_executor, codename="test-codename"
        )

    assert (
        raised.value.brief
        == "Failed to find apt source files in '/etc/apt/sources.list.d/'."
    )


def test_update_apt_sources_dir_sed_error(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error when the sed command fails in the `sources.list.d` directory."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "test", fake_process.any()])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "sed", fake_process.any()])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "find", fake_process.any()],
        stdout="test-output",
    )
    # fail on the third `sed` call
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "sed", fake_process.any()],
        returncode=1,
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_apt_sources(
            executor=fake_executor, codename="test-codename"
        )

    assert (
        raised.value.brief
        == "Failed to update apt source files in '/etc/apt/sources.list.d/'."
    )


def test_update_compatibility_tag(fake_executor, mock_load):
    """`update_compatibility_tag()` should update the instance config."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY, compatibility_tag="test-tag"
    )
    base_config._update_compatibility_tag(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == [
        {
            "content": b"compatibility_tag: test-tag\n",
            "destination": "/etc/craft-instance.conf",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        }
    ]


@pytest.mark.parametrize("status", [True, False])
def test_update_setup_status(fake_executor, mock_load, status):
    """`update_setup_status()` should update the instance config."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    base_config._update_setup_status(
        executor=fake_executor,
        status=status,
    )

    assert fake_executor.records_of_push_file_io == [
        {
            "content": (
                "compatibility_tag: buildd-base-v2\n"
                f"setup: {str(status).lower()}\n".encode()
            ),
            "destination": "/etc/craft-instance.conf",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        }
    ]


def test_ensure_config_compatible_validation_error(fake_executor, mock_load):
    mock_load.side_effect = ValidationError("foo", InstanceConfiguration)

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._ensure_instance_config_compatible(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to parse instance configuration file."
    )


def test_ensure_config_compatible_empty_config_returns_none(fake_executor, mock_load):
    mock_load.return_value = None

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert (
        base_config._ensure_instance_config_compatible(executor=fake_executor) is None
    )


def test_ensure_setup_completed(fake_executor, logs, mock_load):
    """Verify the setup was completed by checking the instance config file."""
    mock_load.return_value = InstanceConfiguration(setup=True)

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert base_config._ensure_setup_completed(executor=fake_executor) is None

    assert "Instance has already been setup." in logs.debug


@pytest.mark.parametrize(
    ("error", "error_message"),
    [
        (
            ValidationError("foo", InstanceConfiguration),
            "failed to parse instance configuration file",
        ),
        (FileNotFoundError, "failed to find instance config file"),
    ],
)
def test_ensure_setup_completed_load_error(
    error, error_message, fake_executor, mock_load
):
    """Raise an error when the instance config cannot be loaded."""
    mock_load.side_effect = error

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError(error_message)


def test_ensure_setup_completed_empty_config(fake_executor, mock_load):
    """Raise an error if the instance config is empty."""
    mock_load.return_value = None

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance config is empty")


@pytest.mark.parametrize("status", [None, False])
def test_ensure_setup_completed_not_setup(status, fake_executor, mock_load):
    """Raise an error if the setup was not completed (setup field is None or False)."""
    mock_load.return_value = InstanceConfiguration(setup=status)

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance is marked as not setup")


@pytest.mark.parametrize(
    "environment",
    [
        None,
        {
            "https_proxy": "http://foo.bar:8081",
            "http_proxy": "http://foo.bar:8080",
        },
    ],
)
@pytest.mark.parametrize("cache_path", [None, pathlib.Path("/tmp")])
def test_warmup_overall(
    environment, fake_process, fake_executor, mock_load, mocker, cache_path
):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=True
    )

    alias = ubuntu.BuilddBaseAlias.JAMMY

    if environment is None:
        environment = ubuntu.BuilddBase.default_command_environment()

    base_config = ubuntu.BuilddBase(alias=alias, environment=environment)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.http=http://foo.bar:8080"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.https=http://foo.bar:8081"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"]
    )

    base_config.warmup(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == []
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == []


def test_warmup_bad_os(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=True
    )
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            """\
            NAME="D.O.S."
            ID=dos
            ID_LIKE=dos
            VERSION_ID="DOS 5.1"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError):
        base_config.warmup(executor=fake_executor)


def test_warmup_bad_instance_config(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )
    base_config.compatibility_tag = "different-tag"

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError):
        base_config.warmup(executor=fake_executor)


@pytest.mark.parametrize("setup", [False, None])
def test_warmup_not_setup(setup, fake_process, fake_executor, mock_load):
    """Raise a BaseConfigurationError if the instance is not setup."""
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=setup
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config.warmup(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance is marked as not setup")


def test_warmup_never_ready(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    for _ in range(3):  # it will called twice until timeout, one extra for safety
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
            stdout="starting",
        )
    base_config._timeout_simple = 0.001

    with pytest.raises(BaseConfigurationError):
        base_config.warmup(executor=fake_executor)


def test_warmup_never_network(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v2", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    for _ in range(3):  # it will called twice until timeout, one extra for safety
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"], returncode=1
        )

    base_config._timeout_simple = 0.001
    with pytest.raises(BaseConfigurationError):
        base_config.warmup(executor=fake_executor)


@pytest.mark.parametrize(
    "hostname",
    [
        "t",
        "test",
        "test1",
        "test-1",
        "1-test",
        "this-is-40-characters-xxxxxxxxxxxxxxxxxx",
        "this-is-63-characters-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    ],
)
def test_set_hostname_unchanged(hostname, logs):
    """Verify hostnames that are already compliant are not changed."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        hostname=hostname,
    )

    assert base_config._hostname == hostname
    assert Exact(f"Using hostname '{hostname}'") in logs.debug


@pytest.mark.parametrize(
    ("hostname", "expected_hostname"),
    [
        # trim away invalid beginning characters
        ("-test", "test"),
        # trim away invalid ending characters
        ("test-", "test"),
        ("test--", "test"),
        ("test1-", "test1"),
        # trim away invalid characters
        ("test$", "test"),
        ("test-!@#$%^&*()test", "test-test"),
        ("$1test", "1test"),
        ("test-$", "test"),
        # this name contains invalid characters so it gets converted, even
        # though it is 63 characters
        (
            "this-is-63-characters-with-invalid-characters-$$$xxxxxxxxxxxxxX",
            "this-is-63-characters-with-invalid-characters-xxxxxxxxxxxxxX",
        ),
        # this name is longer than 63 characters, so it gets converted
        (
            "this-is-64-characters-with-valid-characters-xxxxxxxxxxxxxxxxxxXx",
            "this-is-64-characters-with-valid-characters-xxxxxxxxxxxxxxxxxxX",
        ),
        # trim away away invalid characters and truncate to 63 characters
        (
            "-this-is-64-characters-and-has-invalid-characters-$$$xxxxxxxxxx-",
            "this-is-64-characters-and-has-invalid-characters-xxxxxxxxxx",
        ),
        # ensure invalid ending characters are removed after truncating
        (
            "this-is-64-characters-and-has-a-hyphen-at-character-63-xxxxxxx-x",
            "this-is-64-characters-and-has-a-hyphen-at-character-63-xxxxxxx",
        ),
    ],
)
def test_set_hostname(hostname, expected_hostname, logs):
    """Verify hostname is compliant with hostname naming conventions."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        hostname=hostname,
    )

    assert base_config._hostname == expected_hostname
    assert Exact(f"Using hostname '{expected_hostname}'") in logs.debug


@pytest.mark.parametrize(
    "hostname",
    [
        "",
        "-",
        "$$$",
        "-$-$-",
    ],
)
def test_set_hostname_invalid(hostname):
    """Verify invalid hostnames raise an error."""
    with pytest.raises(BaseConfigurationError) as error:
        ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, hostname=hostname)

    assert error.value == BaseConfigurationError(
        brief=f"failed to create base with hostname {hostname!r}.",
        details="hostname must contain at least one alphanumeric character",
    )


def test_execute_run_default(fake_executor):
    """Default _execute_run behaviour."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    command = ["the", "command"]
    with patch.object(fake_executor, "execute_run") as mock:
        base_config._execute_run(command, executor=fake_executor)

    mock.assert_called_with(
        command, check=True, capture_output=True, text=False, timeout=None
    )


def test_execute_run_options_for_run(fake_executor):
    """Different options to control how run is called."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    command = ["the", "command"]
    with patch.object(fake_executor, "execute_run") as mock:
        base_config._execute_run(
            command,
            executor=fake_executor,
            check=False,
            capture_output=False,
            text=True,
            timeout=None,
        )

    mock.assert_called_with(
        command, check=False, capture_output=False, text=True, timeout=None
    )


def test_execute_run_command_failed_no_verify_network(fake_process, fake_executor):
    """The command failed but network verification was not asked."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor)


@pytest.mark.parametrize("proxy_variable_name", ["HTTPS_PROXY", "https_proxy"])
def test_execute_run_command_failed_verify_network_proxy(
    fake_process, fake_executor, monkeypatch, proxy_variable_name
):
    """The command failed, network verification was asked, but there is a proxy."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    monkeypatch.setenv(proxy_variable_name, "https://someproxy.net:8080/")

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor, verify_network=True)


def test_execute_run_verify_network_run_ok(fake_process, fake_executor):
    """Indicated network verification but process completed ok."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=0)

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    proc = base_config._execute_run(
        command, executor=fake_executor, verify_network=True
    )
    assert proc.returncode == 0


@pytest.mark.usefixtures("stub_verify_network")
def test_execute_run_verify_network_connectivity_ok(fake_process, fake_executor):
    """Network verified after process failure, connectivity ok."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor, verify_network=True)


def test_execute_run_verify_network_connectivity_missing(fake_process, fake_executor):
    """Network verified after process failure, no connectivity."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
        returncode=1,
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    with pytest.raises(NetworkError) as exc_info:
        base_config._execute_run(command, executor=fake_executor, verify_network=True)
    assert isinstance(exc_info.value.__cause__, subprocess.CalledProcessError)


def test_execute_run_bad_check_verifynetwork_combination(fake_executor):
    """Cannot ask for network verification and avoid checking."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    with pytest.raises(RuntimeError):
        base_config._execute_run(
            ["cmd"], executor=fake_executor, check=False, verify_network=True
        )


@pytest.mark.usefixtures("stub_verify_network")
def test_network_connectivity_yes(fake_executor, fake_process):
    """Connectivity is ok."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert base_config._network_connected(executor=fake_executor) is True


def test_network_connectivity_no(fake_executor, fake_process):
    """Connectivity missing."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
        returncode=1,
    )
    assert base_config._network_connected(executor=fake_executor) is False


def test_network_connectivity_timeouts(fake_executor, fake_process):
    """Check that timeout is used.

    This test does not register the fake subprocess with a long wait because to make it
    resilient to CIs it would need a too long waiting.
    """
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    cmd = ["bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"]
    timeout_expired = subprocess.TimeoutExpired(cmd, timeout=5)
    with patch.object(
        fake_executor, "execute_run", side_effect=timeout_expired
    ) as mock:
        assert base_config._network_connected(executor=fake_executor) is False
    mock.assert_called_with(cmd, check=False, capture_output=True, timeout=10)


def test_disable_and_wait_for_snap_refresh_hold_error(fake_process, fake_executor):
    """Raise BaseConfigurationError when the command to hold snap refreshes fails."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._disable_and_wait_for_snap_refresh(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to hold snap refreshes.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


def test_disable_and_wait_for_snap_refresh_wait_error(fake_process, fake_executor):
    """Raise BaseConfigurationError when the `snap watch` command fails."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError) as exc_info:
        base_config._disable_and_wait_for_snap_refresh(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to wait for snap refreshes to complete.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import io
import json
import pathlib
import subprocess
from unittest import mock

import pytest
from craft_providers.errors import details_from_command_error
from craft_providers.multipass import Multipass
from craft_providers.multipass.errors import MultipassError

# Shortcut any calls to time.sleep with pytest-time's instant_sleep
pytestmark = pytest.mark.usefixtures("instant_sleep")

EXAMPLE_INFO = """\
{
    "errors": [
    ],
    "info": {
        "flowing-hawfinch": {
            "disks": {
                "sda1": {
                    "total": "5019643904",
                    "used": "1339375616"
                }
            },
            "image_hash": "c5f2f08c6a1adee1f2f96d84856bf0162d33ea182dae0e8ed45768a86182d110",
            "image_release": "22.04 LTS",
            "ipv4": [
                "10.114.154.206"
            ],
            "load": [
                0.29,
                0.08,
                0.02
            ],
            "memory": {
                "total": 1028894720,
                "used": 152961024
            },
            "mounts": {
            },
            "release": "Ubuntu 22.04.2 LTS",
            "state": "Running"
        }
    }
}
"""

EXAMPLE_LIST = """\
{
    "list": [
        {
            "ipv4": [
            ],
            "name": "manageable-snipe",
            "release": "22.04 LTS",
            "state": "Starting"
        },
        {
            "ipv4": [
                "10.114.154.206"
            ],
            "name": "flowing-hawfinch",
            "release": "22.04 LTS",
            "state": "Running"
        }
    ]
}
"""


@pytest.fixture()
def mock_details_from_process_error():
    details = "<details>"
    with mock.patch(
        "craft_providers.errors.details_from_called_process_error",
        return_value=details,
    ) as mock_details:
        yield mock_details


@pytest.fixture()
def mock_details_from_command_error():
    details = "<details>"
    with mock.patch(
        "craft_providers.errors.details_from_command_error",
        return_value=details,
    ) as mock_details:
        yield mock_details


def test_delete(fake_process):
    fake_process.register_subprocess(["multipass", "delete", "test-instance"])

    Multipass().delete(instance_name="test-instance", purge=False)

    assert len(fake_process.calls) == 1


def test_delete_purge(fake_process):
    fake_process.register_subprocess(
        ["multipass", "delete", "test-instance", "--purge"]
    )

    Multipass().delete(instance_name="test-instance", purge=True)

    assert len(fake_process.calls) == 1


def test_delete_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "delete", "test-instance", "--purge"],
        returncode=1,
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().delete(instance_name="test-instance", purge=True)

    assert exc_info.value == MultipassError(
        brief="Failed to delete VM 'test-instance'.",
        details=mock_details_from_process_error.return_value,
    )


def test_exec(fake_process):
    fake_process.register_subprocess(
        ["multipass", "exec", "test-instance", "--", "sleep", "1"]
    )

    Multipass().exec(command=["sleep", "1"], instance_name="test-instance")

    assert len(fake_process.calls) == 1


def test_exec_error_no_check(fake_process):
    fake_process.register_subprocess(
        ["multipass", "exec", "test-instance", "--", "false"],
        returncode=1,
    )

    proc = Multipass().exec(command=["false"], instance_name="test-instance")

    assert proc.returncode == 1


def test_exec_error_with_check(fake_process):
    fake_process.register_subprocess(
        ["multipass", "exec", "test-instance", "--", "false"],
        returncode=1,
    )

    with pytest.raises(subprocess.CalledProcessError):
        Multipass().exec(command=["false"], instance_name="test-instance", check=True)


def test_info(fake_process):
    fake_process.register_subprocess(
        ["multipass", "info", "test-instance", "--format", "json"], stdout=EXAMPLE_INFO
    )

    data = Multipass().info(instance_name="test-instance")

    assert len(fake_process.calls) == 1
    assert data == json.loads(EXAMPLE_INFO)


def test_info_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "info", "test-instance", "--format", "json"],
        stderr=(
            "info failed: The following errors occurred:\n"
            'instance "foo" does not exist'
        ),
        returncode=1,
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().info(instance_name="test-instance")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to query info for VM 'test-instance'.",
        details=mock_details_from_process_error.return_value,
    )


def test_is_supported_version(fake_process):
    fake_process.register_subprocess(
        ["multipass", "version"], stdout=b"multipass  1.7.0\nmultipassd 1.7.0\n"
    )

    assert Multipass().is_supported_version() is True

    assert len(fake_process.calls) == 1


def test_is_supported_version_false(fake_process):
    fake_process.register_subprocess(
        ["multipass", "version"], stdout=b"multipass  1.4.0\nmultipassd 1.4.0\n"
    )

    assert Multipass().is_supported_version() is False

    assert len(fake_process.calls) == 1


def test_is_supported_version_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(["multipass", "version"], stdout="invalid output")

    with pytest.raises(MultipassError) as exc_info:
        Multipass().is_supported_version()

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Unable to parse version output: b'invalid output'"
    )


def test_launch(fake_process):
    fake_process.register_subprocess(
        ["multipass", "launch", "test-image", "--name", "test-instance"]
    )

    Multipass().launch(image="test-image", instance_name="test-instance")

    assert len(fake_process.calls) == 1


def test_launch_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "multipass",
            "launch",
            "test-image",
            "--name",
            "test-instance",
            "--cpus",
            "4",
            "--memory",
            "8G",
            "--disk",
            "80G",
        ]
    )

    Multipass().launch(
        image="test-image",
        instance_name="test-instance",
        cpus="4",
        mem="8G",
        disk="80G",
    )

    assert len(fake_process.calls) == 1


def test_launch_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "launch", "test-image", "--name", "test-instance"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().launch(instance_name="test-instance", image="test-image")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to launch VM 'test-instance'.",
        details=mock_details_from_process_error.return_value,
    )


def test_list(fake_process):
    fake_process.register_subprocess(
        ["multipass", "list", "--format", "json"], stdout=EXAMPLE_LIST
    )

    vm_list = Multipass().list()

    assert len(fake_process.calls) == 1
    assert vm_list == ["manageable-snipe", "flowing-hawfinch"]


def test_list_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "list", "--format", "json"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().list()

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to query list of VMs.",
        details=mock_details_from_process_error.return_value,
    )


def test_mount(fake_process):
    project_path = pathlib.Path.home() / "my-project"
    fake_process.register_subprocess(
        ["multipass", "mount", str(project_path), "test-instance:/mnt"]
    )

    Multipass().mount(
        source=project_path,
        target="test-instance:/mnt",
        uid_map=None,
        gid_map=None,
    )

    assert len(fake_process.calls) == 1


def test_mount_all_opts(fake_process):
    project_path = pathlib.Path.home() / "my-project"
    fake_process.register_subprocess(
        [
            "multipass",
            "mount",
            str(project_path),
            "test-instance:/mnt",
            "--uid-map",
            "1:2",
            "--uid-map",
            "3:4",
            "--gid-map",
            "5:6",
            "--gid-map",
            "7:8",
        ]
    )

    Multipass().mount(
        source=project_path,
        target="test-instance:/mnt",
        uid_map={"1": "2", "3": "4"},
        gid_map={"5": "6", "7": "8"},
    )

    assert len(fake_process.calls) == 1


def test_mount_error(fake_process, mock_details_from_process_error):
    project_path = pathlib.Path.home() / "my-project"
    fake_process.register_subprocess(
        ["multipass", "mount", str(project_path), "test-instance:/mnt"],
        returncode=1,
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().mount(
            source=project_path,
            target="test-instance:/mnt",
        )

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief=f"Failed to mount {str(project_path)!r} to 'test-instance:/mnt'.",
        details=mock_details_from_process_error.return_value,
    )


def test_start(fake_process):
    fake_process.register_subprocess(["multipass", "start", "test-instance"])

    Multipass().start(instance_name="test-instance")

    assert len(fake_process.calls) == 1


def test_start_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "start", "test-instance"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().start(instance_name="test-instance")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to start VM 'test-instance'.",
        details=mock_details_from_process_error.return_value,
    )


def test_stop(fake_process):
    fake_process.register_subprocess(["multipass", "stop", "test-instance"])

    Multipass().stop(instance_name="test-instance")

    assert len(fake_process.calls) == 1


def test_stop_all_opts(fake_process):
    fake_process.register_subprocess(
        ["multipass", "stop", "test-instance", "--time", "5"]
    )

    Multipass().stop(instance_name="test-instance", delay_mins=5)

    assert len(fake_process.calls) == 1


def test_stop_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "stop", "test-instance"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().stop(instance_name="test-instance")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to stop VM 'test-instance'.",
        details=mock_details_from_process_error.return_value,
    )


def test_transfer(fake_process):
    fake_process.register_subprocess(
        ["multipass", "transfer", "test-instance:/test1", "/test2"]
    )

    Multipass().transfer(source="test-instance:/test1", destination="/test2")

    assert len(fake_process.calls) == 1


def test_transfer_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "transfer", "test-instance:/test1", "/test2"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().transfer(source="test-instance:/test1", destination="/test2")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to transfer 'test-instance:/test1' to '/test2'.",
        details=mock_details_from_process_error.return_value,
    )


def test_transfer_destination_io(fake_process):
    stream = mock.Mock()
    fake_process.register_subprocess(
        ["multipass", "transfer", "test-instance:/test1", "-"], stdout=b"Hello World!\n"
    )

    Multipass().transfer_destination_io(
        source="test-instance:/test1", destination=stream
    )

    assert len(fake_process.calls) == 1
    assert stream.mock_calls == [mock.call.write(b"Hello World!\n")]


def test_transfer_destination_io_chunk_size(fake_process):
    stream = mock.Mock()
    fake_process.register_subprocess(
        ["multipass", "transfer", "test-instance:/test1", "-"], stdout=b"Hello World!\n"
    )

    Multipass().transfer_destination_io(
        source="test-instance:/test1", destination=stream, chunk_size=4
    )

    assert len(fake_process.calls) == 1
    assert stream.mock_calls == [
        mock.call.write(b"Hell"),
        mock.call.write(b"o Wo"),
        mock.call.write(b"rld!"),
        mock.call.write(b"\n"),
    ]


def test_transfer_destination_io_error(fake_process):
    cmd = ["multipass", "transfer", "test-instance:/test1", "-"]
    returncode = 1
    stdout = None
    stderr = b"some stderr"

    fake_process.register_subprocess(
        cmd,
        returncode=returncode,
        stdout=stdout,
        stderr=stderr,
    )

    with io.BytesIO() as stream, pytest.raises(MultipassError) as exc_info:
        Multipass().transfer_destination_io(
            source="test-instance:/test1", destination=stream
        )

    assert exc_info.value == MultipassError(
        brief="Failed to transfer file 'test-instance:/test1'.",
        details=details_from_command_error(
            cmd=cmd, returncode=returncode, stdout=stdout, stderr=stderr
        ),
    )


@mock.patch("subprocess.Popen")
def test_transfer_source_io(mock_popen):
    mock_popen.return_value.__enter__.return_value.returncode = 0

    test_io = io.BytesIO(b"Hello World!\n")

    Multipass().transfer_source_io(
        source=test_io,
        destination="test-instance:/tmp/foo",
    )

    assert mock_popen.mock_calls == [
        mock.call(
            ["multipass", "transfer", "-", "test-instance:/tmp/foo"],
            stdin=-1,
            stderr=-1,
        ),
        mock.call().__enter__(),
        mock.call().__enter__().stdin.write(b"Hello World!\n"),
        mock.call().__enter__().stdin.close(),
        mock.call().__enter__().stderr.read(),
        mock.call().__exit__(None, None, None),
    ]


@mock.patch("subprocess.Popen")
def test_transfer_source_io_chunk_size(mock_popen):
    mock_popen.return_value.__enter__.return_value.returncode = 0

    test_io = io.BytesIO(b"Hello World!\n")

    Multipass().transfer_source_io(
        source=test_io,
        destination="test-instance:/tmp/foo",
        chunk_size=4,
    )

    assert mock_popen.mock_calls == [
        mock.call(
            ["multipass", "transfer", "-", "test-instance:/tmp/foo"],
            stdin=-1,
            stderr=-1,
        ),
        mock.call().__enter__(),
        mock.call().__enter__().stdin.write(b"Hell"),
        mock.call().__enter__().stdin.write(b"o Wo"),
        mock.call().__enter__().stdin.write(b"rld!"),
        mock.call().__enter__().stdin.write(b"\n"),
        mock.call().__enter__().stdin.close(),
        mock.call().__enter__().stderr.read(),
        mock.call().__exit__(None, None, None),
    ]


@mock.patch("subprocess.Popen")
def test_transfer_source_io_error(mock_popen, mock_details_from_command_error):
    mock_popen.return_value.__enter__.return_value.returncode = -1
    test_io = io.BytesIO(b"Hello World!\n")

    with pytest.raises(MultipassError) as exc_info:
        Multipass().transfer_source_io(
            source=test_io,
            destination="test-instance:/tmp/foo",
        )

    assert exc_info.value == MultipassError(
        brief="Failed to transfer file to destination 'test-instance:/tmp/foo'.",
        details=mock_details_from_command_error.return_value,
    )
    assert mock_popen.mock_calls == [
        mock.call(
            ["multipass", "transfer", "-", "test-instance:/tmp/foo"],
            stdin=-1,
            stderr=-1,
        ),
        mock.call().__enter__(),
        mock.call().__enter__().stdin.write(b"Hello World!\n"),
        mock.call().__enter__().stdin.close(),
        mock.call().__enter__().stderr.read(),
        mock.call().__exit__(None, None, None),
    ]


def test_umount(fake_process):
    fake_process.register_subprocess(["multipass", "umount", "test-instance:/mnt"])

    Multipass().umount(mount="test-instance:/mnt")

    assert len(fake_process.calls) == 1


def test_umount_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(
        ["multipass", "umount", "test-instance:/mnt"], returncode=1
    )

    with pytest.raises(MultipassError) as exc_info:
        Multipass().umount(mount="test-instance:/mnt")

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to unmount 'test-instance:/mnt'.",
        details=mock_details_from_process_error.return_value,
    )


def test_wait_until_ready(fake_process, mocker):
    fake_process.register_subprocess(
        ["multipass", "version"],
        returncode=0,
        stdout="multipass  1.6.2\nmultipassd 1.6.3\n",
    )
    multipass = Multipass()
    mocker.patch(
        "craft_providers.const.RETRY_WAIT",
        return_value=0.01,
    )
    multipass_version, multipassd_version = multipass.wait_until_ready()

    assert len(fake_process.calls) == 1

    assert multipass_version == "1.6.2"
    assert multipassd_version == "1.6.3"


@pytest.mark.parametrize("wait_count", [0, 1, 2, 3, 4, 5])
def test_wait_until_ready_with_retries(fake_process, wait_count, mocker):
    for _ in range(0, wait_count):
        fake_process.register_subprocess(
            ["multipass", "version"], returncode=0, stdout="multipass  1.6.2\n"
        )

    fake_process.register_subprocess(
        ["multipass", "version"],
        returncode=0,
        stdout="multipass  1.6.2\nmultipassd 1.6.3\n",
    )

    mocker.patch(
        "craft_providers.const.RETRY_WAIT",
        return_value=0.01,
    )
    multipass = Multipass()
    multipass_version, multipassd_version = multipass.wait_until_ready()

    assert len(fake_process.calls) == wait_count + 1

    assert multipass_version == "1.6.2"
    assert multipassd_version == "1.6.3"


@pytest.mark.parametrize(
    ("time_values", "timeout", "version_calls"),
    [
        ([9.0, 9.5], 0.5, 1),
        ([9.0, 9.4, 9.5], 0.5, 2),
        ([10.0, 20.0, 30.0, 40.0], 30.0, 3),
    ],
)
def test_wait_until_ready_timeout_error(
    fake_process, time_values, timeout, version_calls, mocker
):
    fake_process.register_subprocess(
        ["multipass", "version"],
        returncode=0,
        stdout="multipass  1.6.2\n",
        occurrences=version_calls,
    )
    mocker.patch(
        "craft_providers.const.RETRY_WAIT",
        return_value=0.01,
    )

    with mock.patch("time.time", side_effect=time_values):
        with pytest.raises(MultipassError) as exc_info:
            multipass = Multipass()
            multipass.wait_until_ready(timeout=timeout)

    assert exc_info.value == MultipassError(
        "Timed out waiting for Multipass to become ready."
    )

    assert len(fake_process.calls) == version_calls


@pytest.mark.parametrize(
    ("output", "expected"),
    [
        ("multipass  1.6.2\n", ("1.6.2", None)),
        ("multipass  1.6.2+mac\n", ("1.6.2", None)),
        ("multipass  1.6.2+win\r\n", ("1.6.2", None)),
        ("multipass  1.6.2\nmultipassd 1.6.3\n", ("1.6.2", "1.6.3")),
        ("multipass  1.6.2+mac\nmultipassd 1.6.3+mac\n", ("1.6.2", "1.6.3")),
        ("multipass  1.6.2+win\r\nmultipassd 1.6.3+win\r\n", ("1.6.2", "1.6.3")),
        (
            (
                "multipass  1.6.2+win\r\nmultipassd 1.6.3+win\r\n"
                "some\r\nother\r\nnotice\r\n"
            ),
            ("1.6.2", "1.6.3"),
        ),
    ],
)
def test_version(fake_process, output, expected):
    fake_process.register_subprocess(["multipass", "version"], stdout=output)

    assert Multipass().version() == expected


def test_version_decode_error(fake_process):
    fake_process.register_subprocess(
        ["multipass", "version"], stdout=b"multipass 1.6.2\xd0"
    )

    with mock.patch("locale.getpreferredencoding", return_value="ascii"):
        with pytest.raises(MultipassError) as exc_info:
            Multipass().version()

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to check version.",
        details="Failed to decode output: b'multipass 1.6.2\\xd0'",
    )


def test_version_error(fake_process, mock_details_from_process_error):
    fake_process.register_subprocess(["multipass", "version"], returncode=1)

    with pytest.raises(MultipassError) as exc_info:
        Multipass().version()

    assert len(fake_process.calls) == 1
    assert exc_info.value == MultipassError(
        brief="Failed to check version.",
        details=mock_details_from_process_error.return_value,
    )


def test_generic_run_base():
    """The generic run should check the return code and capture the output."""
    with mock.patch("subprocess.run") as run_mock:
        Multipass()._run(["foo", "1"])
    run_mock.assert_called_with(
        ["multipass", "foo", "1"], check=True, capture_output=True
    )


def test_generic_run_extra_args():
    """Support for extra parameters in the generic run."""
    with mock.patch("subprocess.run") as run_mock:
        Multipass()._run(["foo", "1"], extra="whatever")
    run_mock.assert_called_with(
        ["multipass", "foo", "1"], check=True, capture_output=True, extra="whatever"
    )
